/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __surface_SurfaceCurve_h__
#define __surface_SurfaceCurve_h__

namespace fe
{

/**************************************************************************//**
	@brief Segmented Path Surface

	@ingroup surface

	This surface presumes uniform cv counts.
*//***************************************************************************/
class FE_DL_EXPORT SurfaceCurves: public SurfaceSearchable
{
	public:
	class Impact: public SurfaceSearchable::Impact
	{
		public:
								Impact(void)
								{
#if FE_COUNTED_STORE_TRACKER
									setName("SurfaceCurves::Impact");
#endif
								}
	virtual						~Impact(void)								{}

	virtual	void				draw(const SpatialTransform& a_rTransform,
										sp<DrawI> a_spDrawI,
										const fe::Color* a_pColor) const;

		private:
			void				resolve(void)
								{
									const SurfaceCurves* pSurfaceCurves=
											dynamic_cast
											<const SurfaceCurves*>(
											m_pSurface);
									pSurfaceCurves->resolveImpact(
											sp<ImpactI>(this));
									m_resolved=TRUE;
								}
	};

							SurfaceCurves(void);
virtual						~SurfaceCurves(void)							{}

							//* As Protectable
virtual	Protectable*		clone(Protectable* pInstance=NULL);

							//* As SurfaceI
virtual sp<ImpactI>			sampleImpact(I32 a_triangleIndex,
									SpatialBary a_barycenter,
									SpatialVector a_tangent) const;
virtual	SpatialTransform	sample(I32 a_triangleIndex,SpatialBary a_barycenter,
									SpatialVector a_tangent) const;

virtual SpatialTransform	sample(Vector2 a_uv) const;

virtual	void				draw(const SpatialTransform& a_transform,
									sp<DrawI> a_spDrawI,
									const fe::Color* a_color) const;

	protected:

/*
	class Curve
	{
		public:
							Curve(U32 a_start, U32 a_count):
								m_start(a_start),
								m_count(a_count)							{}

			U32				start(void) const	{ return m_start; }
			U32				count(void) const	{ return m_count; }

		private:
			U32				m_start;
			U32				m_count;
	};
*/

//		Array<Curve>		m_curveArray;

virtual	void				cache(void);

virtual	void				resolveImpact(sp<ImpactI> a_spImpactI) const;

	private:
mutable	CountedPool<Impact>	m_curveImpactPool;
};

} // namespace

#endif /* __surface_SurfaceCurve_h__ */


